/*
 *	Copyright (C) 2000 Hitachi, Ltd.
 *	All rights reserved.
 *	Author: Shingo NISHIOKA
 *		nis@harl.hitachi.co.jp
 */

/* static char rcsid[] = "$Id: getabench.c,v 1.1.1.1 2002/02/22 03:46:31 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/time.h>
#include <sys/uio.h>

#include <assert.h>
#include <errno.h>
#include <math.h>
#include <netdb.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>

#include <geta/nlog.h>
#include <geta/dwut.h>
#include <geta/timer.h>
#include <geta/dw.h>
#include <geta/dci.h>
#include <geta/dcw.h>
#include <geta/dxr.h>
#include <geta/daui.h>
#include <geta/dwwt.h>
#include <geta/dwsh.h>
#include <geta/dwc.h>
#include <geta/dwsched.h>
#include <geta/dwam.h>

#include "tools.h"
#include "tutil.h"

#if HAVE_IEEEFP_H
#include <ieeefp.h>
#endif

static void prepare_test_query(struct syminfo **, int *, int);
static int renso_test(WAM *, struct syminfo *, int);
static void dau_test(char const *, char const *);

static int verbose = 0;

#define	NLOOP		32
/*
#define	NLOOP		128
#define	NLOOP		8192
*/
#define	NQUERIES	16

char *progname = NULL;

static void usage(void);

static void
usage()
{
	fprintf(stderr, "usage: %s [-p|-prob] [-s|-summary #] [-c|-cascade] [-v|-verbose] [-vv|-sexp] [-r|-showresult] dir handle\n", progname);
	fprintf(stderr, "       %s -a|-test-dau dir handle\n", progname);
	exit(1);
}

#if defined USE_MPI
main(argc, argv)
char *argv[];
{
	mpi_main(argc, argv);
	return 0;
}
#endif

#if defined USE_MPI
main0(argc, argv)
#else
main(argc, argv)
#endif
char *argv[];
{
	int showresult = 0;
	int sexp = 0;
	int test_dau = 0;
	int prob_mode = 0;
	int cascade = 0;
	int summary_interval = 0;
	int i, j;
	char *dirname, *handle;
	WAM *w;
	int size;
	struct syminfo *q;
	int nq;
	struct timeval start, elapsed;
	struct gtim *m;
	struct timeval minelap;

	if (progname = rindex(argv[0], '/')) {
		progname++;
	}
	else {
		progname = argv[0];
	}

	for (argc--, argv++; argc>0; argc--, argv++) {
		if (!strcmp(argv[0], "-p") || !strcmp(argv[0], "-prob")) {
			prob_mode = 1;
		}
		else if ((!strcmp(argv[0], "-s") || !strcmp(argv[0], "-summary")) && argc>1) {
			char *e;
			argc--, argv++;
			summary_interval = strtol(argv[0], &e, 10);
			if (*e) {
				usage();
			}
		}
		else if (!strcmp(argv[0], "-c") || !strcmp(argv[0], "-cascade")) {
			cascade = 1;
		}
		else if (!strcmp(argv[0], "-a") || !strcmp(argv[0], "-test-dau")) {
			test_dau = 1;
		}
		else if (!strcmp(argv[0], "-v") || !strcmp(argv[0], "-verbose")) {
			verbose = 1;
		}
		else if (!strcmp(argv[0], "-vv") || !strcmp(argv[0], "-sexp")) {
			sexp = 1;
		}
		else if (!strcmp(argv[0], "-r") || !strcmp(argv[0], "-showresult")) {
			showresult = 1;
		}
		else {
			break;
		}
	}
	if (prob_mode && summary_interval) {
		fprintf(stderr, "sorry, -p and -s should be used exclusively\n");
		return 1;
	}

	if (argc != 2) {
		usage();
	}

#if HAVE_IEEEFP_H && defined fpsetprec
	fpsetprec(FP_PE);
#endif

	nopenlog(progname, LOG_PID, LOG_LOCAL0);

	dirname = argv[0];
	handle = argv[1];

#if defined USE_MPI
	switch (mpi_dwdserv(dirname)) {
	case -1:
		return 1;
	case 0:
		return 0;
	}
#endif

	if (wam_init(dirname)==-1) {
		return 1;
	}

	if (test_dau) {
		dau_test(dirname, handle);
		return 0;
	}

	if (sexp) {
		printf("(%s", handle);
		fflush(stdout);
	}

	if (!(w = wam_open(handle))) {
		nperror("wam_open");
		return 1;
	}

	wam_setopt(w, WAM_OCASCADE, (char *)&cascade);
	wam_setopt(w, WAM_OPROBM, (char *)&prob_mode);
	wam_setopt(w, WAM_OSUMI, (char *)&summary_interval);

	size = wam_size(w, WAM_ROW);

	timerclear(&minelap);
	m = NULL;
	for (j=0; j<4; j++) {
		clear_timers();
		gettimeofday(&start, NULL);
		srandom(1);
		for (i=0; i<NLOOP; i++) {
			nq = NQUERIES;
			prepare_test_query(&q, &nq, size);
			if (renso_test(w, q, nq)==-1) {
				fprintf(stderr, "renso_test failed\n");
				return 1;
			}
		}
		gettimeofday(&elapsed, NULL);

		timersub(&elapsed, &start, &elapsed);

		if (m) {
			struct gtim *g;
			g = get_gtim();
			merge_gtim(g, m, m);
			free(g);
			if (timercmp(&minelap, &elapsed, >)) {
				minelap = elapsed;
			}
		}
		else {
			m = get_gtim();
			minelap = elapsed;
		}
		usleep(100000);
	}

	if (sexp) {
		print_gtim(stdout, &minelap, m);
		printf(" ");
		pkt_d(stdout);

		printf(")\n");
	}

	if (showresult) {
		double d = minelap.tv_sec + minelap.tv_usec/1000000.0;
		printf("%s: %f APS\n", handle, NLOOP/d);
	}

	return 0;
}

static void
prepare_test_query(q, nq, size)
struct syminfo **q;
int *nq;
{
	int i;

	if (!(*q = calloc(*nq, sizeof **q))) {
		nperror("calloc");
		return;
	}

	for (i=0; i<*nq; i++) {
		(*q)[i].id = random()%size+1;
		(*q)[i].TF_d = 1;
		(*q)[i].TF = 1;
		(*q)[i].attr = WSH_OR;
	}
	if (verbose) {
		printf("q = {");
		for (i=0; i<*nq; i++) {
			printf(" %d", (*q)[i].id);
		}
		printf(" }\n");
	}
}

static int
renso_test(w, q, nq)
WAM*w;
struct syminfo *q;
{
	int total;
	int type;

	struct syminfo *d;
	int nd;

	type = WT_SMARTAW;
	nd = 192;

	if (!(d = wsh(q, nq, w, WAM_ROW, type, &nd, &total))) {
		fprintf(stderr, "wsh (1) failed\n");
		return -1;
	}

	free(q);

	if (nd==0) {
		fprintf(stderr, "no intermediate word found\n");
		return -1;
	}

	nq = nd;
	q = d;

	if (verbose) {
		dump_syminfo(stdout, q, nq, NULL);
	}

	type = WT_SMARTWA;
	nd = 512;

	if (!(d = wsh(q, nq, w, WAM_COL, type, &nd, &total))) {
		fprintf(stderr, "wsh (2) failed\n");
		return -1;
	}

	free(q);

	if (verbose) {
		dump_syminfo(stdout, d, nd, NULL);
		fprintf(stdout, "total=%d\n", total);
	}

	free(d);

	return 0;
}

static void
dau_test(dirname, handle)
char const *dirname, *handle;
{
	WAM *f;
	char const *s;
	u_int id;
	ssize_t size;
	ssize_t n;
	char const *cmd;

	cmd = "./sbin/mktitle";

	if (!(f = wam_aux_open(handle, "ttli"))) {
		return;
	}

	n = wam_aux_nentries(f);
	printf("%d\n", n);
	for (id=1; id<=n; id++) {
		errno = 0;
		if ((size = wam_aux_gets(f, id, &s))==-1) {
			nperror("wam_aux_gets");
		}
		else {
			printf("[%s]\n", s);
		}
		if (id==10) {
			break;
		}
	}

	wam_close(f);

	if (!(f = wam_aux_cmd(handle, cmd))) {
		nperror(cmd);
		return;
	}

	for (id=1; id<=n; id++) {
		if ((size = wam_aux_gets(f, id, &s))==-1) {
			nperror("wam_aux_gets");
		}
		printf("[%s]\n", s);
		if (id==10) {
			break;
		}
	}
	wam_close(f);
}
